def printEnergy(atoms, perAtom=True, precision=8, printScript=False):
    import mcol  # https://github.com/mwinokan/MPyTools
    import mout  # https://github.com/mwinokan/MPyTools

    if perAtom:
        epot = atoms.get_potential_energy() / len(atoms)
        ekin = atoms.get_kinetic_energy() / len(atoms)
    else:
        epot = atoms.get_potential_energy()
        ekin = atoms.get_kinetic_energy()

    mout.varOut(
        "E_pot",
        epot,
        unit="eV/atom",
        valCol=mcol.result,
        precision=precision,
        printScript=printScript,
        end=", ",
    )
    mout.varOut(
        "E_kin",
        ekin,
        unit="eV/atom",
        valCol=mcol.result,
        precision=precision,
        printScript=False,
    )



from .group import AtomGroup

import mrich as logger


class System(AtomGroup):
    """Top-level object for molecular systems

    These objects are rarely created by the user,
    instead constructed automatically when parsing a
    coordinate file via amp.parsePDB or otherwise"""

    def __init__(self, name: str):

        super(System, self).__init__(name)

        self.description = None
        self.remarks = []
        self._header_data = []
        self._ssbonds = []

        from .list import NamedList

        self.chains = NamedList()

        self.bondlist = None
        self.box = None

    def autoname_chains(self, verbosity=1):
        """Automatically name chains"""
        import mout

        pro_count = 0
        lig_count = 0
        dna_count = 0
        sol_count = 0

        for i, chain in enumerate(self.chains):
            if chain.type == "PRO":
                if lig_count > 4:
                    mout.warningOut("Too many protein chains!")
                    chain.name = "P"
                else:
                    chain.name = "ABCDE"[pro_count]
                    pro_count += 1
            elif chain.type == "DNA":
                if lig_count > 3:
                    mout.warningOut("Too many DNA chains!")
                    chain.name = "X"
                else:
                    chain.name = "XYZ"[dna_count]
                    dna_count += 1
            elif chain.type == "LIG":
                if lig_count > 1:
                    mout.warningOut("Multiple ligand chains!")
                chain.name = "L"
                lig_count += 1
            elif chain.type == "SOL":
                if sol_count > 1:
                    mout.warningOut("Multiple solvent chains!")
                chain.name = "W"
                sol_count += 1
            elif chain.type == "ION":
                chain.name = chain.residues[0].name[0]

            if verbosity > 0 and chain.name in [c.name for c in self.chains[0:i]]:
                mout.warningOut(
                    f"Ambiguous naming! Multiple chains named {chain.name}!"
                )

    def ssbond_guesser(self, distance: float = 2.15) -> None:
        """Generate SS bonds between SG of CYS according when distance below threshold

        :param distance: minimum distance threshold

        """

        from itertools import combinations
        import numpy as np

        cysteines = [res for res in self.residues if res.name == "CYS"]
        for cys1, cys2 in combinations(cysteines, 2):
            # Get the position of the SG for cysteine 1 and cysteine 2
            sg1 = cys1.get_atom("SG")
            sg2 = cys2.get_atom("SG")
            # Calculate distance between both SGs
            dist = np.linalg.norm(sg1 - sg2)
            # If distance is lower than threshold, append to ssbond
            if dist < distance:
                bond = [
                    {"chain": cys1.chain, "resname": cys1.name, "resid": cys1.number},
                    {"chain": cys2.chain, "resname": cys2.name, "resid": cys2.number},
                    {"sym1": "", "sym2": "", "distance": dist},
                ]
                if any(bond[:2] == check[:2] for check in self._ssbonds):
                    continue
                else:
                    self._ssbonds.append(bond)

    def add_CRYST1(
        self, *, a, b, c, alpha, beta, gamma, space_group, z, debug: bool = False
    ) -> None:
        """Add CRYST1 header record"""

        import mrich

        if debug:
            mrich.var("a", a)
            mrich.var("b", b)
            mrich.var("c", c)
            mrich.var("alpha", alpha)
            mrich.var("beta", beta)
            mrich.var("gamma", gamma)
            mrich.var("space_group", space_group)
            mrich.var("z", z)

        a = float(a)
        b = float(b)
        c = float(c)
        alpha = float(alpha)
        beta = float(beta)
        gamma = float(gamma)
        z = int(z)

        CRYST1 = f"CRYST1{a:9.3f}{b:9.3f}{c:9.3f}{alpha:7.2f}{beta:7.2f}{gamma:7.2f} {space_group:<10} {z:>4}\n"

        if debug:
            mrich.print(CRYST1)

        self._header_data.append(CRYST1)

    def check_indices(self):
        """Print all child Atoms who's indices are incorrect"""

        for index, atom in enumerate(self.atoms):
            if index != atom.index:
                print(index, atom.index, atom.name, atom.residue)

    def fix_indices(self, verbosity=0, exclude=[]):
        """Fix all child Atoms' indices. exclude is a list of residue names to ignore"""
        if verbosity:
            import mout

        for index, atom in enumerate(self.atoms):
            if verbosity == 2 and atom.index != index:
                mout.warningOut(f"Re-indexing atom {atom} (#{atom.index} --> #{index})")
            elif verbosity == 1 and atom.type not in exclude and atom.index != index:
                mout.warningOut(f"Re-indexing atom {atom} (#{atom.index} --> #{index})")
            atom.index = index

        for index, residue in enumerate(self.residues):
            if verbosity == 2 and residue.index != index:
                mout.warningOut(
                    f"Re-indexing residue {residue} (#{residue.index} --> #{index})"
                )
            elif (
                verbosity == 2
                and residue.type not in exclude
                and residue.index != index
            ):
                mout.warningOut(
                    f"Re-indexing residue {residue} (#{residue.index} --> #{index})"
                )
            residue.index = index

            residue.fix_names()
            residue.fix_indices()

    def clear_pdbindices(self):
        """Clear all pdb indices from the system"""
        for atom in self.atoms:
            atom.pdb_index = None

    def clear_atom_numbers(self):
        """Clear all pdb indices from the system"""
        for atom in self.atoms:
            atom._NUMBER = None

    def fix_atomnames(self, verbosity=1):
        """Attempt to fix all child Atom names"""
        import mout

        count = 0
        for index, atom in enumerate(self.atoms):
            if atom.name[0].isnumeric():
                old = atom.name
                new = old[1:] + old[0]
                atom.set_name(new, verbosity=verbosity - 1, element=atom.element)
                if verbosity > 1:
                    mout.out(old + " -> " + new)
                count += 1
        if verbosity > 0 and count != 0:
            mout.warningOut(
                "Fixed " + str(count) + " atom names which appeared to have cycled."
            )

    def add_chain(self, chain):
        """Add a child Chain"""
        from .chain import Chain

        assert isinstance(chain, Chain)
        chain.index = len(self.chains)
        chain.parent = self
        self.chains.append(chain)

    def add_system(self, system, same_chain=False):
        """Merge another system to this one"""

        if same_chain:
            for residue in system.residues:
                self.chains[-1].add_residue(residue.copy())

        else:
            for chain in system.chains:
                self.add_chain(chain.copy())

        self.fix_indices()

    def check_intersection(
        self, system, radius=1, by_residue=True, boolean=False, chain=None
    ):
        """Return a list of indices that are intersecting within a given radius between this system and another."""

        import mout

        # mout.debugOut(f"amp.System.check_intersection({system},radius={radius},by_residue={by_residue},boolean={boolean},chain={chain})")
        import numpy as np

        indices = []

        if by_residue:

            if chain:
                residues = self.get_chain(chain).residues
            else:
                residues = self.residues

            system_CoM = system.CoM(verbosity=0)
            r_system = system.bbox_norm
            num_residues = len(residues)

            # for each residue
            for i, res in enumerate(residues):
                if num_residues > 500 and i % 100 == 0:
                    mout.progress(
                        i,
                        num_residues,
                        prepend="Calculating intersection",
                        append=" of residues checked",
                    )

                residue_CoM = res.CoM(verbosity=0)
                d = np.linalg.norm(residue_CoM - system_CoM)
                r_residue = res.bbox_norm

                if d > r_system + r_residue:
                    continue

                for atom2 in system.atoms:

                    d = np.linalg.norm(residue_CoM - atom2.np_pos)

                    if d > r_system:
                        continue

                    for atom1 in res.atoms:

                        d = np.linalg.norm(atom1.np_pos - atom2.np_pos)

                        if d <= radius:

                            if boolean:
                                return True
                            else:
                                indices.append(res.number)
                                break

            if num_residues > 500:
                mout.progress(
                    num_residues,
                    num_residues,
                    prepend="Calculating intersection",
                    append=" of residues checked. Done.",
                )

        else:

            if chain:
                atoms = self.get_chain(chain).atoms
            else:
                atoms = self.atoms

            for atom1 in self.atoms:
                for atom2 in system.atoms:
                    if np.linalg.norm(atom1.np_pos - atom2.np_pos) <= radius:
                        if boolean:
                            return True
                        else:
                            indices.append(atom1.index)

        if boolean:
            return False
        else:
            return list(set(indices))

    def summary(self, res_limit=10):
        """Print a summary of the System"""
        import mout
        import mcol

        reset = mcol.clear + mcol.bold
        if self.description is not None:
            mout.headerOut(f'\n"{mcol.underline}{self.description}{reset}"')
        mout.headerOut(
            "\nSystem "
            + mcol.arg
            + self.name
            + mcol.clear
            + mcol.bold
            + " contains "
            + mcol.result
            + str(self.num_chains)
            + mcol.clear
            + mcol.bold
            + " chains:"
        )
        for i, chain in enumerate(self.chains):
            mout.headerOut(
                f"Chain[{mcol.arg}{i}{reset}] {mcol.result}{chain.name}{reset} ({mcol.varType}{chain.type}{reset}) [#={mcol.result}{chain.num_atoms}{reset}] =",
                end=" ",
            )

            if chain.type == "PRO":
                names = ""
                for r in chain.residues[:res_limit]:
                    names += r.letter + " "
                if chain.num_residues > res_limit:
                    names += "..."
                mout.out(names)
            else:
                names = ""
                for name in chain.res_names[:res_limit]:
                    names += name + " "
                if chain.num_residues > res_limit:
                    names += "..."
                mout.out(names)
        # mout.varOut("Total Charge",self.charge)

    @property
    def num_chains(self):
        """Number of child Chains (int)"""
        return len(self.chains)

    @property
    def chain_names(self):
        """Get all Chain names (list)"""
        names = []
        for chain in self.chains:
            names.append(chain.name)
        return names

    @property
    def residue_names(self):
        """Get all Chain names (list)"""
        names = []
        for residue in self.residues:
            names.append(residue.name)
        return names

    def rename_atoms(
        self, old: str, new: str, res_filter: str = None, verbosity: int = 2
    ):
        """Rename all matching atoms"""
        import mcol
        import mout

        count = 0
        for residue in self.residues:
            if res_filter is not None and res_filter not in residue.name:
                continue
            for atom in residue.atoms:
                if atom.name == old:
                    count += 1
                    atom.set_name(new, verbosity=verbosity - 1)
        if verbosity > 0:
            if res_filter is None:
                mout.warningOut(
                    "Renamed "
                    + mcol.result
                    + str(count)
                    + mcol.warning
                    + " atoms from "
                    + mcol.arg
                    + old
                    + mcol.warning
                    + " to "
                    + mcol.arg
                    + new
                )
            else:
                mout.warningOut(
                    "Renamed "
                    + mcol.result
                    + str(count)
                    + mcol.warning
                    + " atoms from "
                    + mcol.arg
                    + old
                    + mcol.warning
                    + " to "
                    + mcol.arg
                    + new
                    + mcol.warning
                    + " with res_filter "
                    + mcol.arg
                    + res_filter
                )
        return count

    def rename_residues(self, old: str, new: str, verbosity=2):
        """Rename all matching residues"""
        import mcol
        import mout

        count = 0
        if not isinstance(old, list):
            old = [old]
        for residue in self.residues:
            if residue.name in old:
                # residue.name = new
                residue.rename(new, verbosity=verbosity - 1)
                count += 1
        if verbosity > 0:
            mout.warningOut(
                f"Renamed {mcol.result}{count}{mcol.warning} residues from {mcol.arg}{old}{mcol.warning} to {new}"
            )
        return count

    def get_chain(self, name: str):
        """Get Chain by name"""
        import mout
        import mcol

        for chain in self.chains:
            if chain.name == name:
                return chain
        mout.errorOut(
            "Chain with name " + mcol.arg + name + mcol.error + " not found.",
            fatal=True,
        )

    def remove_chain(self, name, verbosity=1):
        """Delete Chain by name"""
        import mcol
        import mout

        if name not in self.chain_names:
            mout.errorOut(
                "Chain with name " + mcol.arg + name + mcol.error + " not found.",
                fatal=True,
            )
            return

        del_indices = []
        for index, chain in enumerate(self.chains):
            if chain.name == name:
                del_indices.append(index)

        for index in reversed(del_indices):
            c = self.chains[index]
            if verbosity > 0:
                mout.warningOut("Removing chain " + mcol.arg + c.name + str([index]))
            del self.chains[index]

    def remove_heterogens(self, verbosity: int = 1):
        """Remove HETATM entries"""
        import mcol
        import mout

        del_list = []
        atoms = self.atoms
        for index, atom in enumerate(atoms):
            if atom.heterogen:
                del_list.append(index)
        number_deleted = self.remove_atoms(indices=del_list, verbosity=verbosity - 1)
        if verbosity > 0:
            mout.warningOut(
                "Removed "
                + mcol.result
                + str(number_deleted)
                + mcol.warning
                + " heterogens"
            )

    def remove_hydrogens(self, verbosity: int = 1):
        """Remove hydrogen atoms"""
        import mcol
        import mout

        number_deleted = self.remove_atoms(symbols=["H"], verbosity=verbosity - 2)
        if verbosity > 0:
            mout.warningOut(
                "Removed "
                + mcol.result
                + str(number_deleted)
                + mcol.warning
                + " hydrogens"
            )

    def remove_atoms(
        self,
        *,
        names: "list | None" = None,
        indices: "list | None" = None,
        numbers: "list | None" = None,
        symbols: "str | None" = None,
        res_filter: "str | None" = None,
        verbosity: int = 2,
    ) -> int:

        assert (
            names or indices or numbers or symbols
        ), "must supply indices, numbers, names, or symbols"

        import mcol

        mout = logger

        number_deleted = 0

        if indices:
            del_list = indices
            f = lambda x: x.index
        elif numbers:
            del_list = numbers
            f = lambda x: x.number
        elif symbols:
            del_list = symbols
            f = lambda x: x.symbol
        else:
            del_list = names
            f = lambda x: x.name

        for chain in self.chains:
            for residue in chain.residues:
                for index, atom in reversed(list(enumerate(residue.atoms))):
                    if f(atom) in del_list:
                        del residue.atoms[index]
                        number_deleted += 1
                        if verbosity > 1:
                            mout.warning(
                                f"Removed atom {atom.name_number_str} from {residue.name_number_chain_str}"
                            )

        if verbosity > 0:
            mout.var("#atoms deleted", number_deleted)

        return number_deleted

    def remove_residues(
        self,
        *,
        names: "list | None" = None,
        indices: "list | None" = None,
        numbers: "list | None" = None,
        verbosity: int = 2,
        no_summary: bool = False,
    ) -> int:

        assert names or indices or numbers, "must supply indices or numbers"

        import mcol

        mout = logger

        number_deleted = 0

        if indices:
            del_list = indices
            f = lambda x: x.index
        elif numbers:
            del_list = numbers
            f = lambda x: x.number
        else:
            del_list = names
            f = lambda x: x.name

        for c in self.chains:
            for index, residue in reversed(list(enumerate(c.residues))):
                if f(residue) in del_list:
                    del c.residues[index]
                    number_deleted += 1
                    if verbosity > 1:
                        mout.warning(f"Removed residue {residue.name_number_chain_str}")

        if verbosity > 0 and not no_summary:
            mout.var("#residues deleted", number_deleted)

        return number_deleted

    # def atom_names(self, wRes=False, noPrime=False):
    #     """Get all child Atom names (list)"""
    #     names_list = []
    #     for chain in self.chains:
    #         names_list += chain.atom_names(wRes=wRes, noPrime=noPrime)
    #     return names_list

    @property
    def atoms(self):
        """Get all child Atoms (list)"""
        atoms = []
        for chain in self.chains:
            atoms += chain.atoms
        return atoms

    @property
    def QM_indices(self):
        """Return list of Atom indices with the QM flag"""
        index_list = []
        for index, atom in enumerate(self.atoms):
            if atom.QM:
                index_list.append(index)
        return index_list

    @property
    def residues(self):
        """Get all child Residues (list)"""
        residues = []
        for chain in self.chains:
            residues += chain.residues
        return residues

    @property
    def res_names(self):
        """Get all child Residue names (list)"""
        return [res.name for res in self.residues]

    @property
    def num_residues(self):
        """Get number of child Residue (int)"""
        return len(self.residues)

    @property
    def FF_atomtypes(self):
        """Get all child Atom atomtypes (list)"""
        atomtype_list = []
        for chain in self.chains:
            atomtype_list += chain.FF_atomtypes
        return atomtype_list

    @property
    def res_indices(self):
        return [r.index for r in self.residues]

    @property
    def res_numbers(self):
        return [r.number for r in self.residues]

    @property
    def res_names(self):
        return [r.name for r in self.residues]

    def write_CJSON(self, filename, use_atom_types=False, gulp_names=False):
        """Export a CJSON"""
        from .io import writeCJSON

        writeCJSON(filename, self, use_atom_types=use_atom_types, gulp_names=gulp_names)

    def view(self, **kwargs):
        """View the system with ASE"""
        from .gui import view

        view(self, **kwargs)

    def render(self, **kwargs):
        """View the system with py3Dmol"""
        from .py3d import render

        return render(self, **kwargs)

    def auto_rotate(self):
        """Rotate the system into the XY plane"""
        from .manipulate import auto_rotate

        ase_atoms = self.ase_atoms
        ase_atoms = auto_rotate(ase_atoms)
        self.set_coordinates(ase_atoms)

    def align_to(
        self,
        target,
        protein_only=False,
        backbone_only=False,
        heavy_atoms_only=True,
        return_transformations=False,
        verbosity=1,
    ):
        """Align this system to another. use protein_only to align using only the protein, if so the arguments backbone_only and heavy_atoms_only are considered"""

        if protein_only:

            import numpy as np
            from ase.build.rotate import rotation_matrix_from_points
            from .transform import apply_rototranslation, compute_rototranslation_matrix

            # get the protein subsystem
            if backbone_only:
                self_protein = self.protein_backbone
                target_protein = target.protein_backbone

            else:
                self_protein = self.protein_system
                target_protein = target.protein_system

            if heavy_atoms_only:
                self_protein.remove_hydrogens(verbosity=verbosity - 1)
                target_protein.remove_hydrogens(verbosity=verbosity - 1)

            # if the protein subsystems have different residues, use only shared residues
            self_strs = set([r.name_number_str for r in self_protein.residues])
            target_strs = set([r.name_number_str for r in target_protein.residues])

            if self_protein.num_atoms != target_protein.num_atoms:

                assert (
                    len(self_protein.chains) == 1
                ), f"{self.name} {self_protein.chains=}"
                assert (
                    len(target_protein.chains) == 1
                ), f"{target.name} {target_protein.chains=}"

                if verbosity:
                    logger.warning("Proteins have different residues, using common")

                # set arithmetic
                self_remove = self_strs - target_strs
                target_remove = target_strs - self_strs

                # remove not shared
                if self_remove:
                    numbers = [int(s.split()[1]) for s in self_remove]
                    self_protein.remove_residues(
                        numbers=numbers, verbosity=verbosity - 1
                    )

                if target_remove:
                    numbers = [int(s.split()[1]) for s in target_remove]
                    target_protein.remove_residues(
                        numbers=numbers, verbosity=verbosity - 1
                    )

                for a, b in zip(self_protein.residues, target_protein.residues):

                    if a.num_atoms != b.num_atoms:

                        if verbosity:
                            logger.warning(
                                f"{a.name_number_str} has different backbone in self vs target"
                            )
                            logger.warning(f"Pruning alternative sites != A")

                        a.prune_alternative_sites(verbosity=verbosity - 1)
                        b.prune_alternative_sites(verbosity=verbosity - 1)

                        if a.num_atoms != b.num_atoms:
                            a.summary()
                            b.summary()

                            raise Exception()

            # get the atomic positions (ASE style)
            pself = self_protein.ase_atoms.get_positions()
            ptarget = target_protein.ase_atoms.get_positions()

            # translate centeroids to origin
            cself = np.mean(pself, axis=0)
            pself -= cself
            ctarget = np.mean(ptarget, axis=0)
            ptarget -= ctarget

            # get the rotation matrix
            R = rotation_matrix_from_points(pself.T, ptarget.T)

            if return_transformations:
                return cself, ctarget, R

            # get ASE atoms
            atoms = self.ase_atoms

            # apply the rotation matrix
            new_positions = apply_rototranslation(
                atoms.get_positions(), cself, ctarget, R
            )

            atoms.set_positions(new_positions)

            self.set_coordinates(atoms)

            return self

        else:

            from ase.build import minimize_rotation_and_translation

            if isinstance(target, System):
                target = target.ase_atoms
            atoms = self.ase_atoms
            minimize_rotation_and_translation(target, atoms)
            self.set_coordinates(atoms)
            return self

    def apply_transformation(self, matrix):
        """Apply a transformation matrix"""
        from .transform import apply_transformation

        atoms = self.ase_atoms
        new_positions = apply_transformation(atoms.get_positions(), matrix)
        atoms.set_positions(new_positions)
        self.set_coordinates(atoms)

    def translate(self, displacement):
        """Translate the system"""
        self.CoM(shift=displacement, verbosity=0)

    def align_by_pairs(self, target, index_pairs, alt=False):
        """Align the system (i) to the target (j) by consider the vectors:

        a --> b
        a --> c

        where index pairs contains the indices for the three atoms:
        a,b,c in the respective systems:

        index_pairs = [[i_a,j_a],[i_b,j_b],[i_c,j_c]]

        Alternatively you can pass the positions j_a, j_b, j_c as target,
        and index_pairs can be i_a, i_b, i_c.
        """

        assert len(index_pairs) == 3
        import numpy as np

        if isinstance(target, System):
            for pair in index_pairs:
                assert self.atoms[pair[0]].name == target.atoms[pair[1]].name

            pos_0_0 = self.atoms[index_pairs[0][0]].np_pos
            pos_1_0 = self.atoms[index_pairs[1][0]].np_pos
            pos_2_0 = self.atoms[index_pairs[2][0]].np_pos

            pos_0_1 = target.atoms[index_pairs[0][1]].np_pos
            pos_1_1 = target.atoms[index_pairs[1][1]].np_pos
            pos_2_1 = target.atoms[index_pairs[2][1]].np_pos

        else:

            pos_0_0 = self.atoms[index_pairs[0]].np_pos
            pos_1_0 = self.atoms[index_pairs[1]].np_pos
            pos_2_0 = self.atoms[index_pairs[2]].np_pos

            pos_0_1 = target[0]
            pos_1_1 = target[1]
            pos_2_1 = target[2]

        vec = pos_0_1 - pos_0_0
        self.CoM(shift=vec, verbosity=0)

        a = pos_1_0 - pos_0_0
        b = pos_1_1 - pos_0_1
        self.rotate(a, b, pos_0_1)

        a = pos_1_0 - pos_0_0
        a_hat = a / np.linalg.norm(a)

        b = pos_2_0 - pos_0_0
        c = pos_2_1 - pos_0_1
        d = b - np.dot(a, b) * a_hat
        e = c - np.dot(a, c) * a_hat

        d_hat = d / np.linalg.norm(d)
        e_hat = e / np.linalg.norm(e)
        ang = np.arccos(np.clip(np.dot(d_hat, e_hat), -1.0, 1.0))

        if alt:
            self.rotate((ang / np.pi * 180), a, pos_0_1)
        else:
            self.rotate(-(90 - ang / np.pi * 180), a, pos_0_1)

    def guess_names(self, target):
        """Try and set the atom names of the system by looking for
        the closest atom of the same species in the target system"""

        import numpy as np

        positions = [b.np_pos for b in target.atoms]
        species = [b.species for b in target.atoms]

        for a in self.atoms:
            a_pos = a.np_pos
            distances = [
                np.linalg.norm(a_pos - p) if s == a.species else 999
                for s, p in zip(species, positions)
            ]
            index = np.argmin(distances)
            b = target.atoms[index]
            a.set_name(b.name, verbosity=0)

    def rmsd(self, reference):
        """Calculate the RMS displacement between this system and a reference"""

        import numpy as np

        assert len(self.atoms) == len(reference.atoms)
        displacements = np.array(
            [
                np.linalg.norm(a.np_pos - b.np_pos)
                for a, b in zip(self.atoms, reference.atoms)
            ]
        )
        return np.sqrt(np.mean(displacements**2))

    def reorder_atoms(self, reference, map: dict = None):
        new_sys = self.copy()
        for atom in reference.atoms:
            res = self.get_residue(atom.residue, map=map)
            new_sys.add_atom(res.get_atom(atom.name).copy())
        new_sys.fix_indices()
        self = new_sys

    def get_residue(self, name: str, map: dict = None):
        """return residues with matching name"""
        import mout

        if map is not None:
            if name in map.keys():
                name = map[name]
        matches = [r for r in self.residues if r.name == name]
        if len(matches) == 0:
            mout.errorOut(f"No residue found with name {name}")
            return []
        elif len(matches) == 1:
            return matches[0]
        else:
            mout.warning(f"Multiple residues found with name {name}")
            return matches

    def copy(self, disk=False, alt=False):
        """Return a deepcopy of the System"""
        if disk:
            from .io import write, parseGRO

            write(f"__temp__.gro", self, verbosity=0)
            system = parseGRO(f"__temp__.gro", verbosity=0)
            system.name = self.name
            return system
        elif alt:
            if self.num_chains != len(set([str(c) for c in self.chains])):
                import mout

                mout.errorOut(
                    "System has duplicate chain names so this copying method will have incorrect chains!"
                )
            copy_system = System(self.name + " (copy)")
            copy_system.box = self.box
            for atom in self.atoms:
                copy_system.add_atom(atom)
            return copy_system
        else:
            import copy

            return copy.deepcopy(self)

    def subsystem(self, indices, use_pdb_index=True):
        """Extract a subset of the system by the atom indices"""

        new_system = System(self.name + " (filtered)")
        new_system.box = self.box

        for index in indices:

            atom = self.get_atom_by_index(index, use_pdb_index=use_pdb_index)

            # print(index,atom)
            if atom is None:
                continue

            new_system.add_atom(atom)

        return new_system

    def add_atom(self, atom):
        """add an atom to the system"""
        from .chain import Chain
        from .residue import Residue, res_type
        from .list import NamedList

        if isinstance(atom, list) or isinstance(atom, NamedList):
            for a in atom:
                self.add_atom(a)

        else:

            if (
                self.chains
                and atom.chain == self.chains[-1].name
                and res_type(atom.residue) == self.chains[-1].residues[-1].type
            ):
                self.chains[-1].add_atom(atom)
            else:
                chain = Chain(atom.chain)
                chain.add_atom(atom)
                self.add_chain(chain)

    @property
    def children(self):
        return self.chains

    def prune_alternative_sites(self, site="A", verbosity=1):
        """Remove atoms with alternative sites"""
        import mout

        count = 0

        for residue in self.residues:
            count += residue.prune_alternative_sites(site=site, verbosity=verbosity - 1)

        if verbosity > 0 and count > 0:
            mout.warningOut(f"Deleted {count} alternative site atoms")

    def get_protein_interaction_sites(self):
        sites = []
        for chain in self.chains:
            if chain.type != "PRO":
                continue
            for residue in chain.residues:
                sites += residue.interaction_sites
        return sites

    def get_protein_features(self):
        """Get Feature objects for all interaction features on protein atoms in this system. See molparse/rdkit/features"""
        all_features = []
        for chain in self.chains:
            if chain.type != "PRO":
                continue
            for residue in chain.residues:
                features = residue.features
                if not features:
                    mrich.warning("No features", residue.name_number_chain_str)
                    continue
                all_features += residue.features
        return all_features

    def protein_residue_RMSD(self, other, plot=False):

        residue_pairs = []

        data = []

        import numpy as np

        for self_chain in self.protein_system.chains:

            other_chain = other.get_chain(self_chain.name)

            for self_residue in self_chain.residues:
                other_residue = other_chain[
                    f"r{self_residue.name} n{self_residue.number}"
                ]

                distance = np.linalg.norm(self_residue.CoM() - other_residue.CoM())

                data.append(
                    dict(
                        name_number_chain_str=self_residue.name_number_chain_str,
                        self_residue=self_residue,
                        other_residue=other_residue,
                        distance=distance,
                    )
                )

        if plot:
            import plotly.express as px

            fig = px.bar(data, x="name_number_chain_str", y="distance")
            return fig

        return data

    @property
    def protein_system(self):
        sys = self.copy()
        sys.chains = [c for c in sys.chains if c.type == "PRO"]
        return sys

    @property
    def protein_backbone(self):
        sys = self.copy()
        sys.chains = [c for c in sys.chains if c.type == "PRO"]

        for chain in sys.chains:
            for residue in chain.residues:
                residue.remove_sidechain(verbosity=0)

        return sys

    @property
    def ligand_residues(self):
        return [r for r in self.residues if r.type == "LIG"]

    @property
    def ssbonds(self) -> list[list[dict]]:
        """Disulfide bond definition"""
        return self._ssbonds

    def add_ssbond(self, *, cys1: "AminoAcid", cys2: "AminoAcid", sym1, sym2) -> None:
        """Add a disulfide bond definition"""

        from .amino import AminoAcid

        assert isinstance(cys1, Residue)
        assert isinstance(cys2, Residue)
        assert cys1.name == "CYS"
        assert cys2.name == "CYS"

        sg1 = cys1.get_atom("SG")
        sg2 = cys2.get_atom("SG")

        distance = np.linalg.norm(sg1 - sg2)

        self._ssbonds.append(
            [
                {"chain": sg1.chain, "resname": sg1.residue, "resid": sg1.res_number},
                {"chain": sg2.chain, "resname": sg2.residue, "resid": sg2.res_number},
                {"sym1": sym1, "sym2": sym2, "distance": distance},
            ]
        )

    def add_hydrogens(self, pH: float = 7.0, **kwargs) -> "System":
        """Create a protonated copy"""
        from .protonate import protonate

        sys = protonate(self, pH=pH, **kwargs)
        return sys



def get_hbond_pairs(residue1, residue2):
    import mout
    import mcol

    legal = check_legality(residue1, residue2)

    if not legal:
        mout.errorOut(
            "Incompatible DNA residues " + str(residue1) + " & " + str(residue2),
            fatal=True,
        )

    if residue1.name.startswith("DA") or residue1.name == "ADE":
        DA_N1 = residue1.get_atom(name="N1")
        DA_N6 = residue1.get_atom(name="N6")
        DA_H61 = residue1.get_atom(name="H61")
        DA_H62 = residue1.get_atom(name="H62")
        DT_O4 = residue2.get_atom(name="O4")
        DT_N3 = residue2.get_atom(name="N3")
        DT_H3 = residue2.get_atom(name="H3")
        return [[DA_N1, DT_N3, DT_H3], [DT_O4, DA_N6, DA_H61, DA_H62]]
    elif residue1.name.startswith("DT") or residue1.name == "THY":
        DA_N1 = residue2.get_atom(name="N1")
        DA_N6 = residue2.get_atom(name="N6")
        DA_H61 = residue2.get_atom(name="H61")
        DA_H62 = residue2.get_atom(name="H62")
        DT_O4 = residue1.get_atom(name="O4")
        DT_N3 = residue1.get_atom(name="N3")
        DT_H3 = residue1.get_atom(name="H3")
        return [[DA_N1, DT_N3, DT_H3], [DT_O4, DA_N6, DA_H61, DA_H62]]
    elif residue1.name.startswith("DC") or residue1.name == "CYT":
        DC_O2 = residue1.get_atom(name="O2")
        DC_N3 = residue1.get_atom(name="N3")
        DC_N4 = residue1.get_atom(name="N4")
        DC_H41 = residue1.get_atom(name="H41")
        DC_H42 = residue1.get_atom(name="H42")
        DG_N1 = residue2.get_atom(name="N1")
        DG_H1 = residue2.get_atom(name="H1")
        DG_N2 = residue2.get_atom(name="N2")
        DG_H21 = residue2.get_atom(name="H21")
        DG_H22 = residue2.get_atom(name="H22")
        DG_C6 = residue2.get_atom(name="C6")
        DG_O6 = residue2.get_atom(name="O6")
        return [
            [DC_O2, DG_N2, DG_H21, DG_H22],
            [DC_N3, DG_N1, DG_H1],
            [DG_O6, DC_N4, DC_H41, DC_H42],
        ]
    elif residue1.name.startswith("DG") or residue1.name == "GUA":
        DC_O2 = residue2.get_atom(name="O2")
        DC_N3 = residue2.get_atom(name="N3")
        DC_N4 = residue2.get_atom(name="N4")
        DC_H41 = residue2.get_atom(name="H41")
        DC_H42 = residue2.get_atom(name="H42")
        DG_N1 = residue1.get_atom(name="N1")
        DG_H1 = residue1.get_atom(name="H1")
        DG_N2 = residue1.get_atom(name="N2")
        DG_H21 = residue1.get_atom(name="H21")
        DG_H22 = residue1.get_atom(name="H22")
        DG_C6 = residue1.get_atom(name="C6")
        DG_O6 = residue1.get_atom(name="O6")
        return [
            [DC_O2, DG_N2, DG_H21, DG_H22],
            [DC_N3, DG_N1, DG_H1],
            [DG_O6, DC_N4, DC_H41, DC_H42],
        ]

    return False


def check_legality(residue1, residue2):
    if residue1.name.startswith("DA") or residue1.name == "ADE":
        if residue2.name.startswith("DT") or residue2.name == "THY":
            return True
    elif residue1.name.startswith("DT") or residue1.name == "THY":
        if residue2.name.startswith("DA") or residue2.name == "ADE":
            return True
    elif residue1.name.startswith("DG") or residue1.name == "GUA":
        if residue2.name.startswith("DC") or residue2.name == "CYT":
            return True
    elif residue1.name.startswith("DC") or residue1.name == "CYT":
        if residue2.name.startswith("DG") or residue2.name == "GUA":
            return True

    return False


def fix_termini(chain):
    chain.residues[0].make_5ter()
    chain.residues[-1].make_3ter()


def make_5ter(residue):
    import mout

    mout.warningOut(
        "amp.dna.make_5ter is deprecated. Use NucleicAcid.make_5ter instead"
    )

    from ..residue import res_type

    print(residue.name, res_type(residue.name))
    assert res_type(residue.name) == "DNA"

    # Append 5 to residue name
    if not residue.name.endswith("5"):
        residue.name += "5"

    # Remove HTER/H5T
    residue.delete_atom("HTER")
    residue.delete_atom("H5T")
    residue.delete_atom("HO5'")
    residue.delete_atom("OXT")
    residue.delete_atom("O5T")
    residue.delete_atom("O1P")
    residue.delete_atom("O2P")
    residue.delete_atom("OP1")
    residue.delete_atom("OP2")

    # Rename P->H5T
    try:
        residue.get_atom("P").name = "H5T"
    except AttributeError:
        pass


def make_3ter(residue):
    import mout

    mout.warningOut(
        "amp.dna.make_3ter is deprecated. Use NucleicAcid.make_5ter instead"
    )

    from ..residue import res_type

    assert res_type(residue.name) == "DNA"

    # Append 3 to residue name
    if not residue.name.endswith("3"):
        residue.name += "3"

    # Remove HTER/H3T
    residue.delete_atom("O1P3")
    residue.delete_atom("O2P3")
    residue.delete_atom("O3T")
    residue.delete_atom("H3T")
    residue.delete_atom("HO3'")
    residue.delete_atom("HCAP")

    # Rename P3->H3T
    atom = residue.get_atom("P3")
    if atom is not None:
        atom.name = "H3T"


def prep4gmx(system, verbosity=1):
    system.rename_residues("ADE", "DA", verbosity=verbosity)
    system.rename_residues("THY", "DT", verbosity=verbosity)
    system.rename_residues("CYT", "DC", verbosity=verbosity)
    system.rename_residues("GUA", "DG", verbosity=verbosity)

    # Rename DNA Backbone atoms
    system.rename_atoms("OP1", "O1P", res_filter="D", verbosity=verbosity)
    system.rename_atoms("OP2", "O2P", res_filter="D", verbosity=verbosity)
    system.rename_atoms("C7", "C5M", res_filter="D", verbosity=verbosity)

    system.rename_atoms("H5'1", "H5'", res_filter="D", verbosity=verbosity)
    system.rename_atoms("H5'2", "H5''", res_filter="D", verbosity=verbosity)
    system.rename_atoms("H2'1", "H2'", res_filter="D", verbosity=verbosity)
    system.rename_atoms("H2'2", "H2''", res_filter="D", verbosity=verbosity)

    # Deal with DNA termini
    for chain in system.chains:
        if chain.type == "DNA":
            fix_termini(chain)


def get_dna_basepairs(chain1, chain2):
    # checks
    assert chain1.type == "DNA"
    assert chain2.type == "DNA"
    assert len(chain1) == len(chain2)

    return [[a, b] for a, b in zip(chain1.residues, reversed(chain2.residues))]



import mout
import mcol


def parseXVG(
    file,
    xmin=None,
    xmax=None,
    convert_nanometres=True,
    yscale=1.0,
    ylabel=None,
    no_com=False,
    no_ref=False,
    ymin=None,
    ymax=None,
):
    """Parse a GROMACS XVG file and return a molparse.XVG object"""

    mout.out(f"Parsing {mcol.file}{file}{mcol.clear}...")

    header_buffer = []
    xvg = None
    many = []
    multiname = None
    contains_nan = False

    with open(file) as f:
        for line in f:

            # regular comments
            if line.startswith("#"):
                continue

            elif line.startswith("@"):
                header_buffer.append(line[1:].strip())
                continue

            elif line.startswith("&"):
                if many:
                    if xvg.entries:
                        many.append(xvg)
                        many[-1].title = f"[{len(many)}]"
                        xvg = xvg.create_blank_copy()
                else:
                    many = [xvg]
                    multiname = xvg.title
                    many[-1].title = f"[{len(many)}]"
                    xvg = xvg.create_blank_copy()
                continue

            elif not xvg:
                xvg = XVG(ymin=ymin, ymax=ymax)
                if ylabel:
                    xvg.ylabel = ylabel
                xvg.determine_data_shape(
                    header_buffer,
                    line,
                    convert_nanometres,
                    xmin=xmin,
                    xmax=xmax,
                    yscale=yscale,
                )

                xvg.title = f"{file}: {xvg.title}"

                status = xvg.parse_data_line(line)
                contains_nan = contains_nan or not status

            else:
                status = xvg.parse_data_line(line)
                contains_nan = contains_nan or not status

    if many:
        for xvg in many:
            if no_com:
                xvg.remove_com_columns()
            if no_ref:
                xvg.remove_ref_columns()
        many.append(xvg)
        many[-1].title = f"[{len(many)}]"
        if contains_nan:
            mout.warningOut("At least one NaN in XVG!")
        return XVGCollection(many, multiname)
    else:
        if no_com:
            xvg.remove_com_columns()
        if no_ref:
            xvg.remove_ref_columns()
        if contains_nan:
            mout.warningOut("At least one NaN in XVG!")
        return xvg


def clean_label(string):
    """Parses Grace label syntax into LaTeX"""

    substring = string
    chunks = []

    while substring:

        for key in _all_codes:
            if substring.startswith(key):
                chunks.append(key)
                substring = substring[len(key) :]
                break

        else:
            chunks.append(substring[0])
            substring = substring[1:]

    chunks = [c for c in chunks if c not in _ignorable_codes]

    string = ""
    mode = None

    # sub/superscripts
    for chunk in chunks:

        if chunk == "\\s":
            mode = "sub"
            string += r"}_{\text{"
        elif chunk == "\\S":
            mode = "super"
            string += r"}^{\text{"
        elif chunk == "\\N":
            mode = None
            string += r"}}\text{"
        else:
            string += chunk

    string = r"$\text{" + string + "}$"

    return string


class XVG:
    """Class to store data obtained from an XVG"""

    def __init__(self, ymin=None, ymax=None):
        self.entries = 0
        self._minima_indices = None
        self._maxima_indices = None
        self.ymin = ymin
        self.ymax = ymax

    @property
    def dataframe(self):
        import pandas

        return pandas.DataFrame(data=self.columns)

    def determine_data_shape(
        self, header_buffer, demo_line, convert_nanometres, xmin, xmax, yscale=1.0
    ):
        """Use the header strings and an example data line to construct the data shape"""
        self.title = [
            line.lstrip("title ") for line in header_buffer if line.startswith("title")
        ]
        self.xlabel = [
            line.lstrip("xaxis ") for line in header_buffer if line.startswith("xaxis")
        ]
        self.ylabel = [
            line.lstrip("yaxis ") for line in header_buffer if line.startswith("yaxis")
        ]
        self.type = [
            line.lstrip("TYPE ") for line in header_buffer if line.startswith("TYPE")
        ]
        self.series_labels = [
            line for line in header_buffer if line.startswith("s") and "legend" in line
        ]

        if self.series_labels:
            self.series_labels = {
                int(l.split("legend")[0].strip().lstrip("s")): l.split("legend")[1]
                .strip()
                .replace('"', "")
                for l in self.series_labels
            }

        if len(self.title) > 1:
            mout.warningOut("Multiple title headers")
        if len(self.xlabel) > 1:
            mout.warningOut("Multiple xaxis headers")
        if len(self.ylabel) > 1:
            mout.warningOut("Multiple yaxis headers")
        if len(self.type) > 1:
            mout.warningOut("Multiple TYPE headers")

        self.xmin = xmin
        self.xmax = xmax

        self.title = self.title[0]
        self.xlabel = self.xlabel[0].lstrip('label "').rstrip('"')
        self.ylabel = self.ylabel[0].lstrip('label "').rstrip('"')

        self.xlabel = clean_label(self.xlabel)
        self.ylabel = clean_label(self.ylabel)

        self.type = self.type[-1]

        self.xscale = 1.0
        self.yscale = yscale
        if convert_nanometres and "(nm)" in self.xlabel:
            self.xscale = 10.0
            self.xlabel = self.xlabel.replace("(nm)", "(Å)")

        split_line = demo_line.strip().split()

        self.num_columns = len(split_line)

        if self.type == "xy" and self.num_columns == 2:
            self.column_labels = ["x", "y"]
            self.column_types = [float, float]
            self.column_scales = [self.xscale, self.yscale]
            self.columns = {"x": [], "y": []}
        elif self.type == "xy":
            if self.series_labels:
                self.column_labels = ["x"] + [v for _, v in self.series_labels.items()]
            else:
                self.column_labels = ["x"] + [
                    f"y{i + 1}" for i in range(self.num_columns - 1)
                ]
            self.column_scales = [self.xscale] + [self.yscale] * (self.num_columns - 1)
            self.column_types = [float] * self.num_columns
            self.columns = {key: [] for key in self.column_labels}
        elif self.type == "xydy":
            self.column_labels = ["x", "y", "yerr"]
            self.column_scales = [self.xscale, self.yscale, self.yscale]
            self.column_types = [float, float, float]
            self.columns = {"x": [], "y": [], "yerr": []}
        else:
            mout.errorOut(
                f"Unrecognised XVG type: {self.type}, with {self.num_columns} columns",
                fatal=True,
            )

    def parse_data_line(self, line):
        """Parse a line of XVG data"""

        split_line = line.strip().split()

        if len(split_line) < len(self.column_labels):
            mout.warningOut("Skipping incomplete data-line")
            return False

        for value, column, data_type, scale in zip(
            split_line, self.column_labels, self.column_types, self.column_scales
        ):
            if scale is not None:
                value = data_type(value) * scale
            else:
                value = data_type(value)

            if self.xmin and column == "x" and float(self.xmin) > value:
                return True
            if self.xmax and column == "x" and float(self.xmax) < value:
                return True

            self.columns[column].append(value)

        self.entries += 1

        if "nan" in line:
            return False
        else:
            return True

    def plot(self, show=False):
        """Use matplotlib to plot the XVG data"""

        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()
        ax.set_xlabel(self.xlabel)
        ax.set_ylabel(self.ylabel)

        if self.type == "xy" and self.num_columns == 2:
            ax.plot(self.columns["x"], self.columns["y"])

        elif self.type == "xy":
            for key in self.column_labels[1:]:
                ax.plot(self.columns["x"], self.columns[key], label=key)

        elif self.type == "xydy" and self.num_columns == 3:
            ax.errorbar(self.columns["x"], self.columns["y"], yerr=self.columns["yerr"])

        else:
            mout.errorOut(
                f"Unrecognised XVG type (XVG.plot): {self.type}, with {self.num_columns} columns",
                fatal=True,
            )

        if show:
            plt.show()
        return fig, ax

    @property
    def y_columns(self):
        return {k: y for k, y in self.columns.items() if k != "x"}

    @property
    def summable(self):
        if self.type == "xy" and self.num_columns > 2:
            return True
        return False

    def get_summed_trace(self, name=None, color="black", normalised=False):
        import plotly.graph_objects as go

        if self.type == "xy" and self.num_columns > 2:

            summed_data = []
            for i in range(self.entries):

                value = 0
                for key in self.y_columns:
                    value += self.y_columns[key][i]

                summed_data.append(value)

            if normalised:
                maxval = max(summed_data)
                summed_data = [x / maxval for x in summed_data]

            trace = go.Scatter(
                x=self.columns["x"],
                y=summed_data,
                line=dict(color=color),
                name=self.title + " (Summed)",
                mode="lines",
            )

            return trace
        else:
            mout.errorOut(
                f"Unrecognised XVG type (XVG.get_summed_trace): {self.type}, with {self.num_columns} columns",
                fatal=True,
            )

    def get_go_trace(
        self,
        name=None,
        color=None,
        group_from_title=False,
        histogram=False,
        normalised=False,
    ):
        """get the plotly.go trace(s) for the data"""

        import plotly.graph_objects as go

        if normalised:
            normalised = "probability"

        if self.type == "xy" and self.num_columns == 2:
            if histogram:
                ymin = float(self.ymin or min(self.columns["y"]))
                ymax = float(self.ymax or max(self.columns["y"]))
                trace = go.Histogram(
                    x=self.columns["y"],
                    marker_color=color,
                    histnorm=normalised,
                    xbins=dict(
                        start=ymin, end=ymax, size=(ymax - ymin) / float(histogram)
                    ),
                    name=name,
                )
            else:
                trace = go.Scatter(
                    x=self.columns["x"],
                    y=self.columns["y"],
                    line=dict(color=color),
                    name=name,
                    mode="lines",
                )
        elif self.type == "xy":
            traces = []
            for key in self.column_labels[1:]:

                if group_from_title:
                    group = self.title

                else:

                    group = "default"

                    if "pull" in self.title.lower():
                        group = "RC"

                    if "ref" in key:
                        group = "REF"
                    elif "X" in key or "Y" in key or "Z" in key:
                        group = "COM"

                if histogram:
                    ymin = float(self.ymin or min(self.columns[key]))
                    ymax = float(self.ymax or max(self.columns[key]))
                    trace = go.Histogram(
                        x=self.columns[key],
                        marker_color=color,
                        marker_opacity=0.6,
                        name=key,
                        histnorm=normalised,
                        xbins=dict(
                            start=ymin, end=ymax, size=(ymax - ymin) / float(histogram)
                        ),
                        legendgroup=group,
                        legendgrouptitle_text=group,
                    )
                else:
                    trace = go.Scatter(
                        x=self.columns["x"],
                        y=self.columns[key],
                        mode="lines",
                        line=dict(color=color),
                        name=key,
                        legendgroup=group,
                        legendgrouptitle_text=group,
                    )

                traces.append(trace)
            return traces
        elif self.type == "xydy" and self.num_columns == 3:
            if histogram:
                ymin = float(self.ymin or min(self.columns["y"]))
                ymax = float(self.ymax or max(self.columns["y"]))
                trace = go.Histogram(
                    x=self.columns["y"],
                    name=name,
                    histnorm=normalised,
                    marker_color=color,
                )
            else:
                trace = go.Scatter(
                    x=self.columns["x"],
                    y=self.columns["y"],
                    name=name,
                    line=dict(color=color),
                    error_y=dict(type="data", array=self.columns["yerr"], visible=True),
                    mode="lines",
                )
        else:
            mout.errorOut(
                f"Unrecognised XVG type (XVG.plotly): {self.type}, with {self.num_columns} columns",
                fatal=True,
            )

        return trace

    def remove_com_columns(self):
        self.columns = {
            key: column
            for key, column in self.columns.items()
            if not any(["X" in key, "Y" in key, "Z" in key])
        }
        self.column_labels = [
            key
            for key in self.column_labels
            if not any(["X" in key, "Y" in key, "Z" in key])
        ]

    def remove_ref_columns(self):
        self.columns = {
            key: column for key, column in self.columns.items() if not "ref" in key
        }
        self.column_labels = [key for key in self.column_labels if not "ref" in key]

    def calculate_stationary_points(self, column="y"):
        """find any stationary points"""

        import numpy as np
        import scipy.signal as sps

        self.minima_indices = sps.argrelextrema(
            np.array(self.columns[column]), np.less
        )[0]
        self.maxima_indices = sps.argrelextrema(
            np.array(self.columns[column]), np.greater
        )[0]
        self.stationary_column = column

    @property
    def minima_indices(self):
        """List of indices pertaining to minima"""
        if self._minima_indices is None:
            self.calculate_stationary_points()
        return self._minima_indices

    @property
    def maxima_indices(self):
        """List of indices pertaining to maxima"""
        if self._maxima_indices is None:
            self.calculate_stationary_points()
        return self._maxima_indices

    @minima_indices.setter
    def minima_indices(self, a):
        self._minima_indices = a

    @maxima_indices.setter
    def maxima_indices(self, a):
        self._maxima_indices = a

    @property
    def stationary_points(self):
        """build the stationary points dictionary"""
        dictionary = {"minima": [], "maxima": []}
        for i in self.minima_indices:
            dictionary["minima"].append(
                dict(
                    type="min",
                    index=i,
                    x=self.columns["x"][i],
                    y=self.columns[self.stationary_column][i],
                )
            )
        for i in self.maxima_indices:
            dictionary["maxima"].append(
                dict(
                    type="max",
                    index=i,
                    x=self.columns["x"][i],
                    y=self.columns[self.stationary_column][i],
                )
            )
        return dictionary

    def get_stationary_point_trace(self, name=None, column="y", color=None):
        """find any stationary points and make a plotly.go.Scatter trace"""

        import plotly.graph_objects as go

        if self.minima_indices is None or self.maxima_indices is None:
            self.calculate_stationary_points()

        indices = list(self.minima_indices) + list(self.maxima_indices)
        xdata = [self.columns["x"][i] for i in indices]
        ydata = [self.columns[column][i] for i in indices]

        return go.Scatter(
            x=xdata, y=ydata, name=name, mode="markers", marker=dict(color=color)
        )

    def get_closest_value(self, x, column="y", return_x=False):
        from .signal import closest_value

        return closest_value(
            x, self.columns["x"], self.columns[column], return_x=return_x
        )

    def plotly(
        self,
        show=False,
        color=None,
        fig=None,
        group_from_title=False,
        histogram=False,
        summed=False,
        normalised=False,
    ):
        """Use plotly to plot the XVG data"""

        import plotly.graph_objects as go

        if histogram and normalised:
            mout.errorOut("Not supported!", fatal=True)

        if fig is None:
            fig = go.Figure()

        fig.update_layout(
            title=self.title,
            legend=dict(groupclick="toggleitem"),
            font=dict(family="Helvetica Neue", size=18),
        )

        if histogram:
            fig.update_layout(
                xaxis_title=self.ylabel, yaxis_title="Count", barmode="overlay"
            )
            fig.update_traces(opacity=0.75)
        else:
            fig.update_layout(xaxis_title=self.xlabel, yaxis_title=self.ylabel)

        if summed:
            trace = self.get_summed_trace(
                name=self.title, color=color, normalised=normalised
            )
        else:
            trace = self.get_go_trace(
                name=self.title,
                color=color,
                group_from_title=group_from_title,
                histogram=histogram,
                normalised=normalised,
            )

        if isinstance(trace, list):
            for t in trace:
                fig.add_trace(t)
        else:
            fig.add_trace(trace)

        if show:
            fig.show()
        return fig

    def __repr__(self):
        return f"XVG({self.title=}, {self.type=}, {self.xlabel=}, {self.ylabel=}, {self.entries=}, {self.type=}, {self.column_labels=}, {self.column_types=})"

    def create_blank_copy(self):
        """Create a copy of this data structure but with no entries"""

        copy = XVG(ymin=self.ymin, ymax=self.ymax)

        copy.title = self.title
        copy.xlabel = self.xlabel
        copy.ylabel = self.ylabel
        copy.type = self.type
        copy.column_labels = self.column_labels
        copy.column_types = self.column_types
        copy.column_scales = self.column_scales

        copy.xmin = self.xmin
        copy.xmax = self.xmax

        copy.columns = {}
        for key in self.columns:
            copy.columns[key] = []

        copy.num_columns = self.num_columns

        return copy

    def align_ydata(self, method, column="y"):
        """Shift the y-values of all the column to align them according to a function such as min/max or an x-coordinate"""

        import numpy as np
        from .signal import closest_value

        data = self.columns[column]

        if isinstance(method, float):
            ref_value = closest_value(method, self.columns["x"], self.columns[column])
        else:
            ref_value = method(data)

        array = np.array(data)
        array -= ref_value
        self.columns[column] = list(array)

    def smooth(self, column="y", window_length=24, polyorder=3):
        """Smooth a column of data"""
        from scipy.signal import savgol_filter

        self.columns[column] = list(
            savgol_filter(self.columns[column], window_length, polyorder)
        )


class XVGCollection:
    """Class to store many XVG objects"""

    def __init__(self, xvg_list, title):
        self.children = xvg_list
        self.title = title

    def smooth(self, column="y", window_length=24, polyorder=3):
        """Smooth all the children's data"""
        for xvg in self.children:
            xvg.smooth(column, window_length, polyorder)

    def map_xdata(self, before_low, before_high, after_low, after_high):

        scale = (after_high - after_low) / (before_high - before_low)

        for xvg in self.children:

            data = xvg.columns["x"]
            mapped = []

            for x in data:
                mapped.append((x - before_low) * scale + after_low)

            xvg.columns["x"] = mapped

    def plotly(
        self,
        show=False,
        fig=None,
        statistics=False,
        stationary_points=False,
        color=None,
        group_from_title=False,
        no_layout=False,
    ):
        """Use plotly to plot the XVG collection"""

        import plotly.graph_objects as go

        if fig is None:
            fig = go.Figure()

        if not no_layout:
            fig.update_layout(
                title=self.title,
                xaxis_title=self.children[0].xlabel,
                yaxis_title=self.children[0].ylabel,
                legend=dict(groupclick="toggleitem"),
                font=dict(family="Helvetica Neue", size=18),
            )

        if not statistics:

            for xvg in self.children:

                fig.add_trace(
                    xvg.get_go_trace(
                        name=xvg.title, color=color, group_from_title=group_from_title
                    )
                )

                if stationary_points:
                    fig.add_trace(
                        xvg.get_stationary_point_trace(name=xvg.title, color=color)
                    )

        else:

            import numpy as np

            x = self.children[0].columns["x"]
            mean = []
            mean_plus_std = []
            mean_minus_std = []
            mins = []
            maxs = []

            # calculate mean curve
            for i in range(self.children[0].entries):
                this_slice = [xvg.columns["y"][i] for xvg in self.children]

                mins.append(min(this_slice))
                maxs.append(max(this_slice))

                mu = np.mean(this_slice)
                std = np.std(this_slice)

                mean.append(mu)
                mean_plus_std.append(mu + std)
                mean_minus_std.append(mu - std)

            if not color:
                color = "rgb(0,0,0)"  # default is black

            fig.add_trace(
                go.Scatter(
                    x=x,
                    y=maxs,
                    name="max",
                    legendgroup=self.title,
                    legendgrouptitle_text=self.title,
                    line=dict(width=0, color=color),
                )
            )
            fig.add_trace(
                go.Scatter(
                    x=x,
                    y=mins,
                    name="min",
                    fill="tonexty",
                    legendgroup=self.title,
                    legendgrouptitle_text=self.title,
                    line=dict(width=0, color=color),
                    fillcolor=color.replace("rgb", "rgba").replace(")", ",0.15)"),
                )
            )
            fig.add_trace(
                go.Scatter(
                    x=x,
                    y=mean_plus_std,
                    name="mean+std",
                    legendgroup=self.title,
                    legendgrouptitle_text=self.title,
                    line=dict(width=0, color=color),
                )
            )
            fig.add_trace(
                go.Scatter(
                    x=x,
                    y=mean_minus_std,
                    name="mean-std",
                    fill="tonexty",
                    legendgroup=self.title,
                    legendgrouptitle_text=self.title,
                    line=dict(width=0, color=color),
                    fillcolor=color.replace("rgb", "rgba").replace(")", ",0.3)"),
                )
            )
            fig.add_trace(
                go.Scatter(
                    x=x,
                    y=mean,
                    name="mean",
                    legendgroup=self.title,
                    legendgrouptitle_text=self.title,
                    line=dict(color=color, width=4),
                )
            )

            fig.update_layout(showlegend=False)

        if show:
            fig.show()

        return fig

    def align_ydata(self, method):
        """Shift the y-values of all the traces to align them according to a function such as min/max or an x-coordinate"""

        import numpy as np
        from .signal import closest_value

        for xvg in self.children:

            data = xvg.columns["y"]

            if isinstance(method, float):
                ref_value = closest_value(method, xvg.columns["x"], xvg.columns["y"])
            else:
                ref_value = method(data)

            array = np.array(data)
            array -= ref_value
            xvg.columns["y"] = list(array)


### Grace Escape Codes:

# ignore for plotly
_font_codes = {
    "\\f{x}": 'switch to font named "x"',
    "\\f{n}": "switch to font number n",
    "\\f{}": "return to original font",
    "\\x": "switch to Symbol font (same as \\f{Symbol})",
}

# ignore for plotly
_color_codes = {
    "\\R{x}": 'switch to color named "x"',
    "\\R{n}": "switch to color number n",
    "\\R{}": "return to original color",
}

# ignore for plotly
_style_codes = {
    "\\u": "begin underline",
    "\\U": "stop underline",
    "\\o": "begin overline",
    "\\O": "stop overline",
    "\\q": "make font oblique (same as \\l{0.25})",
    "\\Q": "undo oblique (same as \\l{-0.25})",
}

# super/subscripting
_script_codes = {
    "\\s": "begin subscripting (same as \\v{-0.4}\\z{0.71})",
    "\\S": "begin superscripting (same as \\v{0.6}\\z{0.71})",
    "\\N": "return to normal style (same as \\v{}\\t{})",
}

_other_codes = {
    "\\#{x}": 'treat "x" (must be of even length) as list of hexadecimal char codes',
    "\\t{xx xy yx yy}": "apply transformation matrix",
    "\\t{}": "reset transformation matrix",
    "\\z{x}": "zoom x times",
    "\\z{}": "return to original zoom",
    "\\r{x}": "rotate by x degrees",
    "\\l{x}": "slant by factor x",
    "\\v{x}": "shift vertically by x",
    "\\v{}": "return to unshifted baseline",
    "\\V{x}": "shift baseline by x",
    "\\V{}": "reset baseline",
    "\\h{x}": "horizontal shift by x",
    "\\n": "new line",
    "\\Fk": "enable kerning",
    "\\FK": "disable kerning",
    "\\Fl": "enable ligatures",
    "\\FL": "disable ligatures",
    "\\m{n}": "mark current position as n",
    "\\M{n}": "return to saved position n",
    "\\dl": "LtoR substring direction",
    "\\dr": "RtoL substring direction",
    "\\dL": "LtoR text advancing",
    "\\dR": "RtoL text advancing",
    "\\+": "increase size (same as \\z{1.19} ; 1.19 = sqrt(sqrt(2)))",
    "\\-": "decrease size (same as \\z{0.84} ; 0.84 = 1/sqrt(sqrt(2)))",
    "\\T{xx xy yx yy}": "same as \\t{}\\t{xx xy yx yy}",
    "\\Z{x}": "absolute zoom x times (same as \\z{}\\z{x})",
    "\\\\": "print \\",
    "\\n": "switch to font number n (0-9) (deprecated)",
    "\\c": "begin using upper 128 characters of set (deprecated)",
    "\\C": "stop using upper 128 characters of set (deprecated)",
}

_ignorable_codes = _font_codes | _color_codes | _style_codes | _other_codes
_all_codes = _font_codes | _style_codes | _color_codes | _script_codes | _other_codes



def import_checks():
    import importlib

    # tight dependencies
    specs = ["mout", "mcol", "mplot", "mwin"]

    for spec in specs:
        if not importlib.util.find_spec(spec):
            print(f"Fatal Error: Missing '{spec}' package")
            print(
                "Clone the latest version of MPyTools and ensure its in your PYTHONPATH"
            )
            print("https://github.com/mwinokan/MPyTools")
            exit(1)

    # python version
    import sys

    if sys.version_info < (3, 10):
        import mout

        mout.warningOut("Not using python 3.10 or newer! This may lead to errors.")

    # loose dependencies
    dependecies = {
        "ase": [
            "various molecular manipulations and I/O",
            "https://wiki.fysik.dtu.dk/ase/",
        ],
        "numpy": ["countless numerical methods"],
        "scipy": ["curve fitting and analysis "],
        "curses": ["moltree CLI program"],
        "json": ["I/O to the CJSON format"],
        "matplotlib": ["Plotting graphs"],
        "plotly": ["Interactive graphs"],
        # 'imageio': ["creating gif animations (deprecated)"],
    }

    for key in dependecies:
        if not importlib.util.find_spec(key):
            import mout

            mout.warningOut(f"Missing {key} package! This may lead to errors.")
            mout.out(f"MolParse uses {key} for: {dependecies[key][0]}")
            if len(dependecies[key]) > 1:
                mout.out(f"get it from: {dependecies[key][1]}")



import mcol  # https://github.com/mwinokan/MPyTools
import mout  # https://github.com/mwinokan/MPyTools
import copy


class Connectivity:

    def __init__(self, system):
        self.parent_system = system


class Interaction:

    def __init__(self, n_bodies):
        self.n_bodies = n_bodies



def runningAverage(xdata, ydata, averaging_window=5, cutoff_style=None, debug=False):
    import mout

    many = any(isinstance(el, list) for el in ydata)

    if debug:
        mout.varOut("many", many)
        mout.varOut("len(xdata)", len(xdata))
        mout.varOut("len(ydata)", len(ydata))

    if many:

        result_x = []
        result_y = []

        for dataset in ydata:
            x, y = runningAverage(
                xdata,
                dataset,
                averaging_window=averaging_window,
                cutoff_style=cutoff_style,
            )
            result_y.append(y)

        result_x = x

        assert len(ydata) == len(result_y)

    else:

        if cutoff_style is not None:
            mout.errorOut("Not supported yet!", fatal=True)

        result_x = []
        result_y = []

        for i, x in enumerate(xdata):

            result_x.append(x)

            if i == 0:

                y = ydata[0]

            elif i < averaging_window:

                y = sum(ydata[0:i]) / i

            else:

                y = sum(ydata[i - averaging_window : i]) / averaging_window

            result_y.append(y)

    return result_x, result_y


def differentiate(xdata, ydata):
    many = any(isinstance(el, list) for el in ydata)

    result_x = []
    result_y = []

    if many:

        for dataset in ydata:
            x, y = differentiate(xdata, dataset)

            result_y.append(y)

        result_x = x

        assert len(ydata) == len(result_y)

    else:

        for i, x in enumerate(xdata):
            result_x.append(xdata[i])

            if i != 0:

                dy_dx = (ydata[i] - ydata[i - 1]) / (xdata[i] - xdata[i - 1])

                result_y.append(dy_dx)

            else:

                result_y.append(0.0)

    return result_x, result_y



import dash
import numpy as np
import pandas as pd
import plotly.graph_objects as go
from dash import dcc
from dash import html

from scipy.optimize import curve_fit

from dash.dependencies import Input, Output, State

import plotly.graph_objects as go

from IPython.display import display

# selected_points = []

_logfile = None

# from dash import Dash, dcc, html, Input, Output

# app = Dash(__name__)
app = dash.Dash(prevent_initial_callbacks=True)


# @app.callback(
# 	Output(component_id='my-output', component_property='children'),
# 	Input(component_id='graph', component_property='value')
# )
# def update_output_div(input_value):
# 	return f'Output: {input_value}'


def linear(x, a, b):
    return a * x + b


def fitwizard(xdata, ydata, type="linear", mode="markers", logfile=None):
    global _logfile

    _logfile = logfile

    if _logfile is not None:
        f = open(_logfile, "w")
        f.write("# type x_min x_max slope intercept\n")
        f.close()

    # create the plotly graph

    fig = go.FigureWidget(
        [
            go.Scatter(x=xdata, y=ydata, mode=mode),
            go.Scatter(x=None, y=None, mode="lines"),
            go.Scatter(x=None, y=None, mode="markers"),
        ]
    )

    scatter = fig.data[0]

    # colors = ['#a3a7e4'] * 1

    # scatter.marker.color = colors

    # scatter.marker.size = [10] * 100

    fig.layout.hovermode = "closest"

    # fig.update_layout(
    #    updatemenus=[
    #        dict(
    #            type="buttons",
    #            direction="right",
    #            active=0,
    #            x=0.57,
    #            y=1.2,
    #            buttons=list([
    #                dict(label="Select",
    #                     method="update",
    #                     args=[{"dragmode": "select"},
    #                     ]
    #                     ),
    #                dict(label="Zoom",
    #                     method="update",
    #                     args=[{"dragmode": "zoom"},
    #                     ]
    #                     )
    #                ]
    #                )
    #            )
    #        ]
    #        )

    # fig.update_layout(dragmode="select")

    app.layout = html.Div(
        [
            # html.H6("Change the value in the text box to see callbacks in action!"),
            # html.Div([
            dcc.Graph(id="graph", figure=fig),
            # ]),
            # html.Br(),
            # html.Div(id='my-output'),
        ]
    )

    # color = np.zeros(len(ydata), dtype='uint8')
    # colorscale = [[0, '#167b7e'], [1, '#4b3268']]

    # app = dash.Dash(prevent_initial_callbacks=True)
    # app.layout = html.Div([dcc.Graph(figure=build_figure(), id="graph")])

    app.run_server(debug=True)


@app.callback(
    Output("graph", "figure"),
    [Input("graph", "selectedData"), Input("graph", "clickData")],
    [State("graph", "figure")],
)
def linear_fit(selectedData, clickData, fig):
    global _logfile

    selection = None
    # Update selection based on which event triggered the update.
    trigger = dash.callback_context.triggered[0]["prop_id"]
    if trigger == "graph.clickData":
        selection = [point["pointNumber"] for point in clickData["points"]]
    if trigger == "graph.selectedData":
        selection = [point["pointIndex"] for point in selectedData["points"]]

    # Update scatter selection
    fig["data"][0]["selectedpoints"] = selection

    xdata = np.array(fig["data"][0]["x"])
    ydata = np.array(fig["data"][0]["y"])

    sel_x = xdata[selection]
    sel_y = ydata[selection]

    param, covariance = curve_fit(linear, sel_x, sel_y)

    fig["data"][1] = go.Scatter(
        x=sel_x, y=linear(sel_x, param[0], param[1]), mode="lines"
    )

    print(f"Fitting in xrange [{min(sel_x)},{max(sel_x)}]")

    print(f"Slope: {param[0]}")
    print(f"Intercept: {param[1]}")

    if _logfile is not None:
        f = open(_logfile, "a")
        f.write(f"LIN {min(sel_x)} {max(sel_x)} {param[0]} {param[1]}\n")
        f.close()

    return fig



from .residue import Residue
from .group import AtomGroup

ALPHABET = {
    "ATP": "A",
    "TTP": "T",
    "GTP": "G",
    "CTP": "C",
}

LONGNAME = {
    "ATP": "Adenine Tri-Phosphate",
    "TTP": "Thymine Tri-Phosphate",
    "GTP": "Guanine Tri-Phosphate",
    "CTP": "Cytosine Tri-Phosphate",
}


class NucleobaseTriPhosphate(Residue):
    """Class for Nucleobase Tri Phosphate Residue

    These objects should not be created by the user,
    but constructed automatically when parsing a
    coordinate file via amp.parsePDB or otherwise"""

    def __init__(
        self,
        name: str,
        index: int = None,
        number: int = None,
        chain: str = None,
        atoms=None,
    ):
        super(NucleobaseTriPhosphate, self).__init__(name, index, number, chain, atoms)

        self._backbone = None
        self._nucleobase = None

    @property
    def longname(self):
        return LONGNAME[self.name]

    @property
    def letter(self):
        return ALPHABET[self.name]

    @property
    def is_purine(self):
        return self.name in ["GTP", "ATP"]

    @property
    def is_pyrimidine(self):
        return self.name in ["CTP", "TTP"]

    @property
    def type(self):
        """Fixed typing"""
        return "LIG"

    @property
    def backbone(self):
        """Get backbone atoms"""
        if not self._backbone:
            self._backbone = self.get_atom(self.bb_names)
        return self._backbone

    @property
    def nucleobase(self):
        """Get nucleobase atoms"""
        if not self._nucleobase:
            self._nucleobase = self.get_atom(self.nonbb_names)
        return self._nucleobase

    @property
    def bb_names(self):
        return [
            "C1'",
            "H1'",
            "C2'",
            "O2'",
            "H2'",
            "H2''",
            "C3'",
            "H3'",
            "O3'",
            "H3T",
            "HO3'",
            "C4'",
            "H4'",
            "O4'",
            "C5'",
            "H5'",
            "H5''",
            "O5'",
            "PA",
            "O1A",
            "O2A",
            "O3A",
            "PB",
            "O1B",
            "O2B",
            "O3B",
            "PG",
            "O1G",
            "O2G",
            "O3G",
        ]

    @property
    def nonbb_names(self):
        """Get nucleobase atoms"""
        return [n for n in self.atom_names() if n not in self.bb_names]

    def flip(self):

        import mout
        import mcol

        mout.header(
            f"Flipping {mcol.arg}{self.longname} {mcol.clear}{mcol.bold}({mcol.arg}{self.name_number_str}{mcol.clear}{mcol.bold})"
        )

        nucleobase_group = AtomGroup.from_any(
            f"{self.name}.nucleobase", self.nucleobase
        )

        if self.is_purine:
            source = nucleobase_group.atoms["N9"][0]
            target = nucleobase_group.atoms["O6"][0]

        vector = target - source

        nucleobase_group.rotate(180, vector, source.np_pos)

        for a1, a2 in zip(self.nucleobase, nucleobase_group.atoms):
            a1.position = a2.position

    def mutate(self, newname, show=False):
        """Mutate this nucleic triphosphate to another"""

        import mout
        import mcol

        mout.header(
            f"Mutating {mcol.arg}{self.longname}{mcol.clear + mcol.bold} --> {mcol.arg}{LONGNAME[newname]}{mcol.clear + mcol.bold} ({mcol.result}{self.letter}{self.number}{ALPHABET[newname]}{mcol.clear + mcol.bold})"
        )

        if self.name == newname:
            mout.warning("Skipping mutation with self == target!")
            return

        NEW_ANALOGUE = {
            "ATP": "DA",
            "TTP": "DT",
            "GTP": "DG",
            "CTP": "DC",
        }

        import os

        amp_path = os.path.dirname(__file__)

        from .io import parse

        if newname in ["ATP", "TTP"]:
            ref_sys = parse(f"{amp_path}/ref/AT_FLAT.pdb", verbosity=0)
        elif newname in ["GTP", "CTP"]:
            ref_sys = parse(f"{amp_path}/ref/GC_FLAT.pdb", verbosity=0)
        else:
            mout.error("Unsupported target mutation")

        rem = [c for c in ref_sys.chain_names if c not in NEW_ANALOGUE[newname][-1]][0]
        ref_sys.remove_chain(rem, verbosity=0)
        ref_sys.fix_indices()
        ref_res = ref_sys.residues[0]

        if self.is_purine and ref_res.is_purine:

            # get atoms to align by
            names = ["N9", "C6", "C2"]
            start = ref_res.get_atom(names)
            target = self.get_atom(names)

            # align the system
            extra = ref_sys.align_by_posmap([start, target])

            # delete intersecting atoms
            ref_res.delete_atom("H9", verbosity=0)

        elif self.is_pyrimidine and ref_res.is_pyrimidine:

            # get atoms to align by
            names = ["C6", "C4", "C2"]
            start = ref_res.get_atom(names)
            target = self.get_atom(names)

            # align the system
            extra = ref_sys.align_by_posmap([start, target])

            # shift to connect to the backbone
            ref_res.translate(self.get_atom("N1") - ref_res.get_atom("N1"))

            # delete intersecting atoms
            ref_res.delete_atom("H1", verbosity=0)

        elif self.is_purine and ref_res.is_pyrimidine:
            mout.warning("Purine --> Pyrimidine")

            # get atoms to align by
            start = ref_res.get_atom(["C6", "C4", "C2"])
            target = self.get_atom(["C4", "C6", "C2"])

            # align the system
            extra = ref_sys.align_by_posmap([start, target])

            # shift to connect to the backbone
            ref_res.translate(self.get_atom("N9") - ref_res.get_atom("N1"))

            # delete intersecting atoms
            ref_res.delete_atom("H1", verbosity=0)

        elif self.is_pyrimidine and ref_res.is_purine:
            mout.warning("Pyrimidine --> Purine")

            # get atoms to align by
            start = ref_res.get_atom(["C4", "C6", "C2"])
            target = self.get_atom(["C6", "C4", "C2"])

            # align the system
            extra = ref_sys.align_by_posmap([start, target])

            # shift to connect to the backbone
            ref_res.translate(self.get_atom("N1") - ref_res.get_atom("N9"))

            # delete intersecting atoms
            ref_res.delete_atom("H9", verbosity=0)

        self.name = newname

        if show:
            view_sys = ref_sys.copy()
            for atom in self.atoms:
                view_sys.add_atom(atom)
            view_sys.plot3d(extra, 1.0)

        # delete intersecting atoms
        self.delete_atom(self.nonbb_names, verbosity=0)

        # add the reference nucleobase
        for atom in ref_res.atoms:
            self.add_atom(atom)

    def oxidise(self):

        assert self.name == "GTP"

        import numpy as np
        from .atom import Atom

        import mout
        import mcol

        mout.header(
            f"Oxidising {mcol.arg}{self.longname}{mcol.header} --> {mcol.result}{mcol.bold}8-Oxo-7,8-dihydroguanine triphosphate (OGTP)"
        )

        # H8 --> O8
        self.get_atom("H8").name = "O8"

        # +H7
        N7 = self.get_atom("N7")
        N9 = self.get_atom("N9")
        C4 = self.get_atom("C4")
        vec = N7 - (N9 + C4) / 2
        vec /= np.linalg.norm(vec)
        atom = Atom("H7")
        atom.position = N7 + vec
        self.add_atom(atom)

        # +H2
        N2 = self.get_atom("N2")
        C2 = self.get_atom("C2")
        N3 = self.get_atom("N3")
        vec = np.cross(N2 - C2, N3 - C2)
        vec /= np.linalg.norm(vec)
        atom = Atom("H2")
        atom.position = C2 + vec
        self.add_atom(atom)

        # +H3
        C6 = self.get_atom("C6")
        vec = N3 - C6
        vec /= np.linalg.norm(vec)
        atom = Atom("H3")
        atom.position = N3 + vec
        self.add_atom(atom)

        self.name = "OGTP"



from rdkit import Chem


def strip_quotes(val):
    if val[0] == '"':
        val = val[1:]
    if val[-1] == '"':
        val = val[:-1]
    return val


BOND_TYPES = {
    "single": Chem.rdchem.BondType.SINGLE,
    "double": Chem.rdchem.BondType.DOUBLE,
    "triple": Chem.rdchem.BondType.TRIPLE,
    "SINGLE": Chem.rdchem.BondType.SINGLE,
    "DOUBLE": Chem.rdchem.BondType.DOUBLE,
    "TRIPLE": Chem.rdchem.BondType.TRIPLE,
    "aromatic": Chem.rdchem.BondType.AROMATIC,
    "deloc": Chem.rdchem.BondType.SINGLE,
    "SING": Chem.rdchem.BondType.SINGLE,
    "DOUB": Chem.rdchem.BondType.DOUBLE,
    "TRIP": Chem.rdchem.BondType.TRIPLE,
}



from .modify import runningAverage
from .modify import differentiate

# from .2d import peakFinder

from .io import parseDat
from .io import csv_strip

from .peaks import peakFinder, closest_index, closest_value

# the fitwizard should be loaded only as needed via: from ase.signal

# import socket
# if not 'eureka' in socket.gethostname():
# 	if not 'login' in socket.gethostname():
# 		if not 'node' in socket.gethostname():
# 			from .fitwizard import fitwizard



from .hijack import FakeAtoms, FakeSurfaceCalculator
import mout
import mcol

from ase.optimize import BFGS
from ase.units import kcal, mol
from ase.calculators.singlepoint import SinglePointCalculator


class NDFES(object):
    def __init__(self, outdir, n_coords):
        mout.debugHeader("NDFES.__init__()")
        self._pmf = None
        self._meta = None
        self._calc = None
        self.outdir = outdir
        self.checkpoint = f"{outdir}/checkpoint"
        self.metafile = f"{outdir}/metafile"
        self.n_coords = n_coords
        super(NDFES, self).__init__()

    def run(self, method="vfep", binlength=0.2, temperature=300, script="eureka.sh"):
        mout.debugHeader("NDFES.run()")

        mout.varOut("metafile", self.metafile, valCol=mcol.file)
        mout.varOut("checkpoint", self.checkpoint, valCol=mcol.file)
        mout.varOut("method", f"{method}", valCol=mcol.arg)
        mout.varOut("temperature", f"{temperature}", unit="K", valCol=mcol.arg)
        mout.varOut("script", f"{script}", valCol=mcol.file)

        import os

        amp_path = os.environ["MWAMPPATH"]
        script = f"{amp_path}/molparse/ndfes/{script}"

        length_str = ""
        for i in range(self.n_coords):
            length_str += f"-w {binlength} "

        mout.headerOut("Running ndfes...")
        os.system(
            f"{script} --{method} -c {self.checkpoint} {length_str} {self.metafile} -t {temperature} > {self.outdir}/ndfes.log"
        )

    @property
    def pmf(self):
        if self._pmf is None:
            self.generate_pmf_object()
        return self._pmf

    @property
    def meta(self):
        if self._meta is None:
            self.generate_pmf_object()
        return self._meta

    @property
    def calc(self):
        return self._calc

    @calc.setter
    def calc(self, arg):
        self._calc = arg

    def energy_along_path(self, rcs):
        energies = []
        for xs in rcs:
            e = self.pmf([xs])
            # if e > 1000:
            # 	energies.append(None)
            # else:
            # 	energies.append(e*(kcal/mol))
            energies.append(e * (kcal / mol))
        return energies

    def generate_pmf_object(self):
        mout.debugHeader("NDFES.generate_pmf_object()")
        try:
            import ndfes
        except:
            mout.errorOut(
                "Missing ndfes library. Try sourcing $MWSHPATH/load_ndfes.sh",
                fatal=True,
            )

        meta = ndfes.Metafile(self.metafile)
        pmf = ndfes.interpolator([self.checkpoint], meta)

        self._meta = meta
        self._pmf = pmf

    def plot_2d_fes(self, dims):
        mout.debugHeader("NDFES.plot_2d_fes()")
        try:
            import ndfes
        except:
            mout.errorOut(
                "Missing ndfes library. Try sourcing $MWSHPATH/load_ndfes.sh",
                fatal=True,
            )

        mout.varOut("checkpoint", f"{self.checkpoint}", valCol=mcol.file)

        ndgrid = ndfes.NDUniformRegularGrid()
        for d in dims:
            ndgrid.add_dim_using_delta(d[0], d[1], d[2])

        pts = ndgrid.get_list_of_tuples()

        vals = self.pmf(pts)

        # Print the PMF to a file
        fh = open(f"{self.outdir}/2dfes.dat", "w")
        for pt, val in zip(pts, vals):
            # if pt[0] == -1.0: fh.write("\n")

            # collapse onto 2d

            if val > 1000:
                continue

            fh.write("%15.5f %15.5f %20.10e\n" % (pt[0] - pt[1], pt[2] - pt[3], val))

        fh.close()

    def plot_window_pmf(self, i1, i2):
        import matplotlib.pyplot as plt
        import numpy as np

        mout.debugHeader("NDFES.plot_window_pmf()")

        try:
            import ndfes
        except:
            mout.errorOut(
                "Missing ndfes library. Try: "
                + mcol.file
                + "source $MWSHPATH/load_ndfes.sh",
                fatal=True,
            )

        cs = self.meta.get_centers()

        # reactant free energy
        self.reactant = cs[i1]
        mout.varOut(
            f"Reactant Free Energy (window {i1})",
            self.pmf([self.reactant])[0],
            unit="kcal/mol",
        )

        # product free energy
        self.product = cs[i2]
        mout.varOut(
            f"Product Free Energy (window {i2})",
            self.pmf([self.product])[0],
            unit="kcal/mol",
        )

        cs = cs[0:5]

        # reshape into 1d

        vals = self.pmf(cs)

        x = []
        y = []

        mout.headerOut("Window Centres & Energies:")

        for c, v in zip(cs, vals):
            x.append(c[0] - c[1] + c[2] - c[3])
            print(c, v)
            if v < 1000:
                y.append(v * (kcal / mol))
            else:
                y.append(None)

        fig, ax = plt.subplots()

        plt.scatter(x, y)

        plt.savefig(f"{self.outdir}/window_pmf.png")

        plt.close()

    def find_minima(self, start, fmax=0.01, optimizer=BFGS, suppress_warnings=False):
        mout.debugHeader("NDFES.find_minima()")

        # mout.varOut("start",list(start))

        atoms = FakeAtoms(start)

        # print("1",atoms.get_positions())

        self.calc = FakeSurfaceCalculator(self.pmf, suppress_warnings=suppress_warnings)

        atoms.calc = self.calc

        # print("2",atoms.get_positions())

        # from ase.optimize import BFGS

        dyn = optimizer(atoms)

        # print("3",atoms.get_positions())

        dyn.run(fmax=fmax)

        mout.varOut("Post-BFGS RC's", atoms.rcs)

    def constraint_pair(self, index, locut=0.5, hicut=3.0, k=20):
        from ase.constraints import Hookean

        print(index, locut, hicut)
        if locut > hicut:
            c1 = Hookean(a1=index, a2=(1, 0, 0, -locut), k=k)
            c2 = Hookean(a1=index, a2=(-1, 0, 0, hicut), k=k)
        else:
            c1 = Hookean(a1=index, a2=(1, 0, 0, -hicut), k=k)
            c2 = Hookean(a1=index, a2=(-1, 0, 0, locut), k=k)
        return [c1, c2]

    def find_path(
        self,
        start,
        final,
        n_images=10,
        fmax=0.01,
        traj=None,
        optimizer=BFGS,
        suppress_warnings=True,
        constraints=None,
        constrain_final=False,
    ):

        assert n_images > 2

        if traj is None:
            traj = f"{self.outdir}/neb_rcs.traj"

        from ase.neb import NEB
        from ase.optimize import MDMin

        if constrain_final:
            from ase.constraints import FixAtoms

        start = FakeAtoms(start)
        final = FakeAtoms(final)

        mout.varOut("NEB Start", start.rcs)
        mout.varOut("NEB Final", final.rcs)

        images = [start]
        for i in range(n_images - 2):
            image = start.copy()
            image.calc = FakeSurfaceCalculator(
                self.pmf, suppress_warnings=suppress_warnings
            )
            image.set_constraints(constraints[0], constraints[1], constraints[2])
            images.append(image)
        images.append(final)

        neb = NEB(images)
        neb.interpolate()

        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()

        interpolated_rcs = [[p[0] for p in a.get_positions()] for a in images]

        fig, ax = plt.subplots()
        plt.plot(self.energy_along_path(interpolated_rcs))
        plt.savefig(f"{self.outdir}/neb_interpolated.png")
        plt.close()

        dyn = optimizer(neb, trajectory=traj)
        dyn.run(fmax=fmax)


def pullx2rco(infile, output, n_coords, skip=1):
    """convert a pullx to an rco file"""

    import mout
    import os

    mout.out(f"{infile} --> {os.path.basename(output)}")

    f_input = open(infile, "r")
    f_output = open(output, "w")

    mins = [None for i in range(n_coords)]
    maxs = [None for i in range(n_coords)]

    for line in f_input:

        if line[0] in ["#", "@"]:
            count = 0
            continue

        if count % skip != 0:
            count += 1
            continue

        # the new RCO should contain
        # time, N*(restraint)

        split_line = line.split()

        time = float(split_line[0])
        RCs = []

        for i in range(n_coords):
            x = float(split_line[1 + 8 * i]) * 10
            RCs.append(x)
            if mins[i] is None or x < mins[i]:
                mins[i] = x
            if maxs[i] is None or x > maxs[i]:
                maxs[i] = x

        f_output.write(f"{time:.4f}")
        for x in RCs:
            f_output.write(f" {x:.6f}")

        f_output.write("\n")

        count += 1

    # return reference information
    window_centres = []
    for i in range(n_coords):
        window_centres.append(float(split_line[2 + 8 * i]))

    f_input.close()
    f_output.close()

    return window_centres, mins, maxs


def prepare_metafile(outdir, winkeys, n_coords, force, skip=1):
    import mout
    import mcol
    import os

    mout.debugHeader("prepare_metafile()")

    os.system(f"mkdir -p {outdir}")

    metafile = open(f"{outdir}/metafile", "w")

    mout.varOut("metafile", f"{outdir}/metafile", valCol=mcol.file)
    mout.varOut("force", f"{force}", unit="kcal/mol/Length^2", valCol=mcol.arg)

    all_mins = [None for i in range(n_coords)]
    all_maxs = [None for i in range(n_coords)]

    mout.headerOut("Creating amber-style RCO's")
    for key in winkeys:

        # create the 'amber-style' RCO's
        centres, mins, maxs = pullx2rco(
            f"{key}/pullx.xvg", f"{outdir}/{key}.rco", n_coords, skip=skip
        )

        for i, (this_min, this_max) in enumerate(zip(mins, maxs)):
            if all_mins[i] is None or this_min < all_mins[i]:
                all_mins[i] = this_min
            if all_maxs[i] is None or this_max > all_maxs[i]:
                all_maxs[i] = this_max

        # add to the metafile
        # the file should contain
        # RCO_path, N*(RN_window_centre, RN_force)

        metafile.write(f"{outdir}/{key}.rco")
        for x in centres:
            metafile.write(f" {x * 10:.6f} {force:.2f}")
        metafile.write("\n")

    for i in range(n_coords):
        mout.varOut(f"RC{i + 1} range: ", f"({all_mins[i]} - {all_maxs[i]})", unit="Å")


## use scipy to find minima and maxima in the 4D PMF



import numpy as np
from ase.calculators.calculator import Calculator
from ase import Atoms
import mout
import mcol


class NaNEncounteredInPmfError(Exception):
    pass


class FakeAtoms(Atoms):
    """This class creates an object resembling ase.Atoms that encodes any number (N) of reaction coordinates into the x-positions of ase.Atom objects. Use this class together with FakeSurfaceCalculator to perform optimisations inside an N-dimensional PMF.

    Initialise as:

    FakeAtoms(rcs=[...],name=...)

    - rcs is a list of floats representing a vector in reaction-coordinate space
    - name is an optional string naming this point in RC-space

    """

    def __init__(self, rcs=None, name=None, **kwargs):
        mout.debugOut(f"FakeAtoms.__init__({rcs=},{name=},...)")

        self.name = name

        if rcs is not None:

            symbols = len(rcs) * "H"

            positions = []

            for r in rcs:
                positions.append((r, 0.0, 0.0))

            Atoms.__init__(self, symbols, positions=positions, **kwargs)

        else:

            Atoms.__init__(self, **kwargs)

    def set_constraints(self, locut=0.5, hicut=3.0, k=20, additional=None):

        del self.constraints

        from ase.constraints import Hookean

        constraints = []

        for i in range(self.__len__()):
            constraints.append(Hookean(a1=i, a2=(1, 0, 0, -hicut), k=k))
            constraints.append(Hookean(a1=i, a2=(-1, 0, 0, locut), k=k))

        # constraints.append(Hookean(a1=i,a2=(-1,0,0,-hicut),k=k))
        # constraints.append(Hookean(a1=i,a2=(1,0,0,locut),k=k))

        if additional is not None:
            constraints.append(additional)

        self.set_constraint(constraints)

    def positions_from_rcs(self, rcs):
        positions = []
        for r in rcs:
            positions.append((r, 0.0, 0.0))
        self.positions = positions

    @property
    def rcs(self):
        return [p[0] for p in self.positions]

    def set_positions(self, newpositions, apply_constraint=False):
        newpositions = np.array([np.array([p[0], 0.0, 0.0]) for p in newpositions])
        self.set_array("positions", newpositions, shape=(3,))


class FakeSurfaceCalculator(Calculator):
    """

    Calculator to hijack ASE optimisation algorithms.

    Initialise as:

    FakeSurfaceCalculator(pmf=function())

    pmf is a N-dimensional free energy surface. Should be callable with the same number of arguments as reaction coordinates described in the FakeAtoms object.

    Optimisation example:

    atoms = FakeAtoms(rcs=[1.0,1.0])

    def pmf(x,y):
        ...

    calc = FakeSurfaceCalculator(pmf=pmf)

    atoms.set_calculator(calc)

    from ase.optimize import BFGS

    dyn = BFGS(atoms=atoms)

    dyn.run()

    """

    implemented_properties = ["energy", "forces"]

    # default_parameters = {'delta': 0.01}

    nolabel = True

    def __init__(
        self, pmf, delta=0.001, suppress_warnings=False, restoring_force=10, **kwargs
    ):
        mout.debugOut(f"FakeSurfaceCalculator.__init__(delta={delta})")
        self._pmf = pmf
        self.delta = delta
        self._history = []
        self._last_ok = None
        self._restoring_force = restoring_force
        self._warnings = not suppress_warnings
        Calculator.__init__(self, **kwargs)

    def pmf(self, arg):
        from ase.units import kcal, mol

        e = self._pmf(arg)

        if np.isnan(e):
            mout.errorOut(f"NaN detected in PMF! {arg=}")
            raise NaNEncounteredInPmfError

        return self._pmf(arg)

    @property
    def history(self):
        return self._history

    @history.setter
    def history(self, arg):
        self._history = arg

    def calculate(
        self,
        atoms=None,
        properties=["energy", "forces"],
        system_changes=["positions", "numbers", "cell", "pbc"],
    ):
        Calculator.calculate(self, atoms, properties, system_changes)

        self.energy = 0.0

        xs = np.array([p[0] for p in self.atoms.get_positions()])

        energy = self.pmf([xs])
        mout.debugOut(f"xs: {xs}, pmf(xs): {energy}")

        self.history.append(xs)

        forces = []

        y = self.pmf([xs])

        self._last_ok = [x for x in xs]

        # partial derivatives
        for i in range(len(self.atoms)):
            h = np.array([0.0 for i in range(len(self.atoms))])
            h[i] = self.delta
            y_h = self.pmf([xs + h])

            dx = (y_h - y) / self.delta
            forces.append(-dx[0])

        mout.debugOut(f"forces: {forces}")

        forces = np.array([np.array([f, 0.0, 0.0]) for f in forces])

        ### convert forces to correct units

        self.results["energy"] = energy
        self.results["forces"] = forces
